#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <conio.h>
#include <time.h>

#include <windows.h>		// for windows specific keywords in ftd2xx.h
#include "ftd2xx.h"			// Header file for ftd2xx.lib
#include "ft2232BDM.h"
#include "mpc8xxbdm.h"
#include "mpc8xxmisc.h"


FT_HANDLE ftHandle;				// Handle to access the chip
FT_DEVICE ftDevice;
HANDLE WinHandle;                // Win32 handle for (Windows)Event Creation
FT_STATUS ftStatus;                // To store the status value

uint8_t GPIO_ACBUS_data_direction = 0;
uint8_t GPIO_ACBUS_data_value = 0;

uint8_t GPIO_ADBUS_data_direction = 0;
uint8_t GPIO_ADBUS_data_value = 0;

int verbose = 0;

void set_verbose (int v)
{
    verbose = v;
}

int get_verbose ()
{
    return verbose;
}

int open_FT232_USB_comport (int index, uint32_t BaudRate)
{
//    uint32_t   dwEventMask;              // Event Mask for setting conditions for triggering the event

    ftStatus = FT_Open(index, &ftHandle);
    if (ftStatus != FT_OK)
    {
        printf("E1) - Error in opening FT232_USB handle \n");
        return 0;
    }

    ftStatus = FT_SetBaudRate(ftHandle, BaudRate); //Setting up the BaudRate
    if (ftStatus != FT_OK)
    {
       printf("E2) - Error in setting BaudRate to %d\n", BaudRate);
       return 0;
    }
/*
    ftStatus = FT_SetDataCharacteristics (ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
    if (ftStatus != FT_OK)
    {
        printf("E3) - Error in setting Data Format \n");
        return 0;
    }

    ftStatus = FT_SetFlowControl (ftHandle, FT_FLOW_NONE, 0, 0);
    if (ftStatus != FT_OK)
    {
        printf("E4) - Error in setting Flow Control \n");
        return 0;
    }

    WinHandle = CreateEvent (NULL, FALSE, FALSE, "FT232Event");
    if (WinHandle == INVALID_HANDLE_VALUE)
    {
        printf("E5) - ERROR in Creating Event\n");
        return 0;
    }

    dwEventMask = FT_EVENT_RXCHAR; //Receiving a Character should trigger the event

    ftStatus = FT_SetEventNotification (ftHandle, dwEventMask, WinHandle);
    if (ftStatus != FT_OK)
    {
        printf("E6) - FT_SetEventNotification() Failure \n");
        return 0;
    }
*/
    return 1;
}

void printStatus()
{
    printf("status = %d\n", (int)ftStatus);
}

int show_FTDI_devices (int *numberOfUnitsOfWantedType, int list_of_wanted_units_indexes[])
{
    FT_HANDLE ftHandleTemp;
    uint32_t numDevs;
    uint32_t Flags;
    uint32_t ID;
    uint32_t Type;
    uint32_t LocId;
    char SerialNumber[16];
    char Description[64];
    int i;
    int numDevsWanted = 0;

    ftStatus = FT_CreateDeviceInfoList((LPDWORD)&numDevs);
    if (ftStatus != FT_OK)
        return 0;
//    else
//        printf("Number of devices is %d\n", (int)numDevs);


    if (numDevs > 0)
    {
        for (i=0; i<numDevs; i++)
        {

//            printf ("device #%d:\n", i);

            ftStatus = FT_GetDeviceInfoDetail(i, (LPDWORD)&Flags, (LPDWORD)&Type, (LPDWORD)&ID, (LPDWORD)&LocId, SerialNumber, Description, &ftHandleTemp);
            if (ftStatus == FT_OK)
            {
/*                printf("Dev 0:\n");
                printf(" Flags=0x%x\n", (int)Flags);
                printf(" Type=0x%x\n", (int)Type);
                printf(" ID=0x%x\n", (int)ID);
                printf(" LocId=0x%x\n", (int)LocId);
                printf(" SerialNumber=%s\n", SerialNumber);
                printf(" Description=%s\n", Description);
                printf(" ftHandle=0x%x\n", (int)ftHandleTemp);
*/
                if ((strcmp("FT2232H MiniModule A", Description)==0) || (strcmp("USB <-> Serial Converter A", Description)==0) || (strcmp("Quad RS232-HS A", Description)==0)
                    || (strcmp("TDS3000_2ndLIFE A", Description)==0))
                {
//                    printf("that's what we want\n");
                    if (numDevsWanted<32)
                    {
                        list_of_wanted_units_indexes[numDevsWanted] = i;
                        numDevsWanted++;
                    }
                }
            }
        }
    }

    for (i=0; i<numDevsWanted; i++)
    {
        printf ("FT2232H MiniModule A or USB <-> Serial Converter A #%d at index %d\n", i, list_of_wanted_units_indexes[i]);
    }
    *numberOfUnitsOfWantedType = numDevsWanted;
    return (int)numDevs;
}

int GetGPIOinput (uint8_t HighByte)
{
    DWORD dwNumBytesToSend = 0;
    DWORD dwNumBytesSent = 0;
    DWORD dwNumBytesToRead = 0;
    DWORD dwNumBytesRead = 0;
    BYTE byOutputBuffer[MAX_BUFFER]; // Buffer to hold MPSSE commands and data to be sent to the FT2232H
    BYTE byInputBuffer[MAX_BUFFER]; // Buffer to hold data read from the FT2232H

    if (HighByte)
        byOutputBuffer[dwNumBytesToSend++] = 0x83;
    else
        byOutputBuffer[dwNumBytesToSend++] = 0x81;
    // Get data bits - returns state of pins,either input or output on low byte of MPSSE
    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);

    do
    {
        ftStatus = FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
    } while ((dwNumBytesToRead < 1) && (ftStatus == FT_OK));


    dwNumBytesToSend = 0; // Reset output buffer pointer

    ftStatus |= FT_Read(ftHandle, &byInputBuffer, dwNumBytesToRead, &dwNumBytesRead);
    if ((ftStatus != FT_OK) & (dwNumBytesToRead != 1))
    {
        printf("Error - GPIO cannot be read\n");
        FT_SetBitMode(ftHandle, 0x0, 0x00);
        // Reset the port to disable MPSSE
        FT_Close(ftHandle); // Close the USB port
        return -1; // Exit with error
    }
    return (int)byInputBuffer[0];
}

int Get_GPIO_HighByte_input ()
{
    return GetGPIOinput (1);
}

int Get_GPIO_LowByte_input ()
{
    return GetGPIOinput (0);
}

void SetGPIO_output (uint8_t HighByte)
{
    DWORD dwNumBytesToSend = 0;
    DWORD dwNumBytesSent = 0;
    BYTE byOutputBuffer[MAX_BUFFER]; // Buffer to hold MPSSE commands and data to be sent to the FT2232H
    if (HighByte)
    {
        byOutputBuffer[dwNumBytesToSend++] = 0x82;
        byOutputBuffer[dwNumBytesToSend++] = GPIO_ACBUS_data_value;
        byOutputBuffer[dwNumBytesToSend++] = GPIO_ACBUS_data_direction;
    }
    else
    {
        byOutputBuffer[dwNumBytesToSend++] = 0x80;
        byOutputBuffer[dwNumBytesToSend++] = GPIO_ADBUS_data_value;
        byOutputBuffer[dwNumBytesToSend++] = GPIO_ADBUS_data_direction;
    }

    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
}

void Set_GPIO_HighByte_output ()
{
    SetGPIO_output (1);
}

void Set_GPIO_LowByte_output ()
{
    SetGPIO_output (0);
}

void LoopBackOnOrOff (int LoopBackOn)
{
    DWORD dwNumBytesToSend = 0; // Reset output buffer pointer
    DWORD dwNumBytesSent = 0;
    uint8_t byOutputBuffer[2];
    // Disable or Enable internal loop-back
    if (LoopBackOn)
        byOutputBuffer[dwNumBytesToSend++] = 0x84;
    else
        byOutputBuffer[dwNumBytesToSend++] = 0x85;

    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
    // Send off/on the loopback command
}

void Set_ClockDivisor (DWORD dwClockDivisor)
{
    DWORD dwNumBytesToSend = 0; // Reset output buffer pointer
    DWORD dwNumBytesSent = 0;
    uint8_t byOutputBuffer[4];
    // Set TCK frequency
    // TCK = 60MHz /((1 + [(1 +0xValueH*256) OR 0xValueL])*2)
    byOutputBuffer[dwNumBytesToSend++] = 0x86;
    //Command to set clock divisor
    byOutputBuffer[dwNumBytesToSend++] = dwClockDivisor & 0xFF;
    //Set 0xValueL of clock divisor
    byOutputBuffer[dwNumBytesToSend++] = (dwClockDivisor >> 8) & 0xFF;
    //Set 0xValueH of clock divisor
    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
    // Send off the clock divisor commands
}

int SynchorniseMPSSE ()
{
    // -----------------------------------------------------------
    // Synchronize the MPSSE by sending a bogus opcode (0xAA),
    // The MPSSE will respond with "Bad Command" (0xFA) followed by
    // the bogus opcode itself.
    // -----------------------------------------------------------
    DWORD dwNumBytesToSend = 0; // Reset output buffer pointer
    DWORD dwNumBytesSent = 0;
    uint8_t byOutputBuffer[4];
    DWORD dwNumBytesToRead = 0; // Number of bytes available to read in the driver's input buffer
    DWORD dwNumBytesRead = 0; // Count of actual bytes read - used with FT_Read
    BYTE byInputBuffer[MAX_BUFFER]; // Buffer to hold data read from the FT2232H
    DWORD dwCount = 0; // General loop index
    int bCommandEchod = 0;

    byOutputBuffer[dwNumBytesToSend++] = 0xAA;//'\xAA';
    //Add bogus command xAA to the queue
    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
    // Send off the BAD commands
    Sleep(50); // Wait for all the USB stuff to complete and work

    do
    {

        do
        {
            ftStatus = FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
            // Get the number of bytes in the device input buffer
        } while ((dwNumBytesToRead == 0) && (ftStatus == FT_OK));
        //or Timeout

        if (dwNumBytesToRead > MAX_BUFFER)
            dwNumBytesToRead = MAX_BUFFER;

        ftStatus = FT_Read(ftHandle, byInputBuffer, dwNumBytesToRead, &dwNumBytesRead);
        //Read out the data from input buffer

        for (dwCount = 0; dwCount < dwNumBytesRead - 1; dwCount++)
        //Check if Bad command and echo command received
        {
            if ((byInputBuffer[dwCount] == 0xFA) && (byInputBuffer[dwCount+1] == 0xAA))
            {
                bCommandEchod = 1;
                break;
            }
        }
    } while (!bCommandEchod);

    return bCommandEchod;
}

int get_ADBUS (uint8_t mask)
{
    if (Get_GPIO_LowByte_input () & mask)
        return 1;
    return 0;
}

void set_clear_ADBUS (int value, uint8_t mask)
{
    if (value)
        GPIO_ADBUS_data_value |= mask;
    else
        GPIO_ADBUS_data_value &= ~mask;
    Set_GPIO_LowByte_output ();
}

void setDSDI ()
{
    set_clear_ADBUS (1, MASK_DSDI);
}

void clearDSDI ()
{
    set_clear_ADBUS (0, MASK_DSDI);
}

void setDSCK ()
{
    set_clear_ADBUS (1, MASK_DSCK);
}

void clearDSCK ()
{
    set_clear_ADBUS (0, MASK_DSCK);
}

void assertSRESET()
{
    GPIO_ADBUS_data_direction |= MASK_SRESET; // output
    set_clear_ADBUS (0, MASK_SRESET);
}

void releaseSRESET ()
{
    GPIO_ADBUS_data_direction &= ~MASK_SRESET; // input so it's open collector
    set_clear_ADBUS (1, MASK_SRESET);
}

void assertHRESET()
{
    GPIO_ADBUS_data_direction |= MASK_HRESET; // output
    set_clear_ADBUS (0, MASK_HRESET);
}

void releaseHRESET ()
{
    GPIO_ADBUS_data_direction &= ~MASK_HRESET; // input so it's open collector
    set_clear_ADBUS (1, MASK_HRESET);
}

int getFRZ ()
{
    return get_ADBUS (MASK_FRZ);
}

int getDSDO ()
{
    return get_ADBUS (MASK_DSDO);
}

void inout34or9bits (uint8_t out_first3bits, uint32_t out_next32bits, int len, uint8_t *in_first3bits, uint32_t *in_next32bits)
{
    DWORD dwNumBytesToSend = 0; // Reset output buffer pointer
    DWORD dwNumBytesSent = 0;
    uint8_t byOutputBuffer[16];
    DWORD dwNumBytesToRead = 0; // Number of bytes available to read in the driver's input buffer
    DWORD dwNumBytesRead = 0; // Count of actual bytes read - used with FT_Read
    DWORD expected_bytes_back = 0;

    BYTE byInputBuffer[10]; // Buffer to hold data read from the FT2232H
    do
    {
        ftStatus |= FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
        ftStatus |= FT_Read(ftHandle, byOutputBuffer, dwNumBytesToRead, &dwNumBytesRead);
    } while (dwNumBytesToRead);

    byOutputBuffer[dwNumBytesToSend++] = 0x33;
    byOutputBuffer[dwNumBytesToSend++] = 0x02;  // 3 bits
    byOutputBuffer[dwNumBytesToSend++] = (out_first3bits | 4) << 5;

    if (len == 7)
    {
        byOutputBuffer[dwNumBytesToSend++] = 0x33;
        byOutputBuffer[dwNumBytesToSend++] = 0x06;  // 7 bits
        expected_bytes_back = 2;
        byOutputBuffer[dwNumBytesToSend++] = out_next32bits << 1;
    }
    else
    {
        byOutputBuffer[dwNumBytesToSend++] = 0x31;
        byOutputBuffer[dwNumBytesToSend++] = 0x03;  // 4 bytes = 32 bits
        byOutputBuffer[dwNumBytesToSend++] = 0x00;
        byOutputBuffer[dwNumBytesToSend++] = out_next32bits >> 24;
        byOutputBuffer[dwNumBytesToSend++] = out_next32bits >> 16;
        byOutputBuffer[dwNumBytesToSend++] = out_next32bits >> 8;
        byOutputBuffer[dwNumBytesToSend++] = out_next32bits >> 0;
        expected_bytes_back = 5;
    }
    ftStatus |= FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
    do
        ftStatus |= FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
    while (dwNumBytesToRead != expected_bytes_back);

    ftStatus |= FT_Read(ftHandle, byInputBuffer, dwNumBytesToRead, &dwNumBytesRead);

    *in_first3bits = byInputBuffer[0] & 3;
    if (*in_first3bits & 2)
        *in_next32bits = byInputBuffer[1];
    else
        *in_next32bits = (((uint32_t)byInputBuffer[1])<<24) + (((uint32_t)byInputBuffer[2])<<16)+ (((uint32_t)byInputBuffer[3])<<8)+ (((uint32_t)byInputBuffer[4])>>0);
}

int inout1bit (uint8_t out_bit)
{
    DWORD dwNumBytesToSend = 0; // Reset output buffer pointer
    DWORD dwNumBytesSent = 0;
    uint8_t byOutputBuffer[16];
    DWORD dwNumBytesToRead = 0; // Number of bytes available to read in the driver's input buffer
    DWORD dwNumBytesRead = 0; // Count of actual bytes read - used with FT_Read
    DWORD expected_bytes_back = 0;

    BYTE byInputBuffer[10]; // Buffer to hold data read from the FT2232H

    do
    {
        ftStatus |= FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
        ftStatus |= FT_Read(ftHandle, byOutputBuffer, dwNumBytesToRead, &dwNumBytesRead);
    } while (dwNumBytesToRead);

    byOutputBuffer[dwNumBytesToSend++] = 0x33;
    byOutputBuffer[dwNumBytesToSend++] = 0x00;  // 1 bits
    if (out_bit)
        byOutputBuffer[dwNumBytesToSend++] = 0x80;
    else
        byOutputBuffer[dwNumBytesToSend++] = 0;

    expected_bytes_back = 1;
    ftStatus |= FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);

    do
        ftStatus |= FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
    while (dwNumBytesToRead != expected_bytes_back);

    ftStatus |= FT_Read(ftHandle, byInputBuffer, dwNumBytesToRead, &dwNumBytesRead);

    if (byInputBuffer[0] & 1)
        return 1;
    else
        return 0;
 }


int find_FTDI_FT2232H_OR_FT4232H_A_index ()
{
    int list_of_wanted_units_indexes[32];
    int numberOfUnitsOfWantedType=0;
    int i, n, index;

    n = show_FTDI_devices (&numberOfUnitsOfWantedType, list_of_wanted_units_indexes);
    if (!numberOfUnitsOfWantedType)
    {
        printf ("no FTDI FT2232H/FT4232H A devices found, exiting\n");
        return -1;
    }

    index = list_of_wanted_units_indexes[0];
    printf ("%d devices found\n", numberOfUnitsOfWantedType);
    if (numberOfUnitsOfWantedType > 1)
    {
        for (i=0; i<numberOfUnitsOfWantedType; i++)
        {
            printf ("Compatible FT2232H/FT4232H A #%d at index %d\n", i, list_of_wanted_units_indexes[i]);
        }

        printf ("use which device index? (0..%d):", n-1);
        scanf ("%d", &index);
        if ((index<0) || (index>(n-1)))
        {
            printf ("%d is out of range, exiting\n", index);
            return -1;
        }
    }
    return index;
}

int open_FT2232_for_BDM (int index)
{
    uint32_t deviceID;
    char SerialNumber[16];
    char Description[64];

    DWORD dwNumBytesToSend = 0; // Index to the output buffer
    DWORD dwNumBytesSent = 0; // Count of actual bytes sent - used with FT_Write
    DWORD dwNumBytesToRead = 0; // Number of bytes available to read in the driver's input buffer
    DWORD dwNumBytesRead = 0; // Count of actual bytes read - used with FT_Read

//    DWORD dwClockDivisor = 0x05DB; // Value of clock divisor, SCL Frequency = 60/((1+0x05DB)*2) (MHz) = 20khz
//    DWORD dwClockDivisor = 29; // Value of clock divisor, SCL Frequency = 60/((1+29)*2) (MHz) = 1.00 MHz

    BYTE byInputBuffer[MAX_BUFFER]; // Buffer to hold data read from the FT2232H
    BYTE byOutputBuffer[MAX_BUFFER]; // Buffer to hold MPSSE commands and data to be sent to the FT2232H

    if(!open_FT232_USB_comport (index, 3000000))
    {
       printf ("\n error USB connecting to FTDI unit\n");
       return -1;
    }

    ftStatus = FT_GetDeviceInfo(ftHandle, &ftDevice, (LPDWORD)&deviceID, SerialNumber, Description, NULL);
    if (ftStatus == FT_OK)
    {
        if (ftDevice == FT_DEVICE_232H)
            printf ("device is FT232H\n"); // device is FT232H
        else if (ftDevice == FT_DEVICE_4232H)
            printf ("device is FT4232H\n"); // device is FT4232H
        else if (ftDevice == FT_DEVICE_2232H)
            printf ("device is FT2232H\n"); // device is FT2232H
        else if (ftDevice == FT_DEVICE_232R)
            printf ("device is FT232R\n"); // device is FT232R
        else if (ftDevice == FT_DEVICE_2232C)
            printf ("device is FT2232C/L/D\n"); // device is FT2232C/L/D
        else if (ftDevice == FT_DEVICE_BM)
            printf ("device is FTU232BM\n"); // device is FTU232BM
        else if (ftDevice == FT_DEVICE_AM)
            printf ("device is FT8U232AM\n"); // device is FT8U232AM
        else
            printf ("device is UNKNOWN\n"); // unknown device (this should not happen!)
            // deviceID contains encoded device ID
            // SerialNumber, Description contain 0-terminated strings

        printf ("serial_nr = %s\n", SerialNumber);
        printf ("description = %s\n", Description);

    }
    else
    {
        printf ("FT_GetDeviceType FAILED\n");// FT_GetDeviceType FAILED!
        return -1;
    }


    ftStatus = FT_SetBitMode (ftHandle, 0b11111111,1);

    printStatus();
    ftStatus = FT_ResetDevice(ftHandle);
    ftStatus |= FT_GetQueueStatus(ftHandle, &dwNumBytesToRead);
    // Get the number of bytes in the FT2232H receive buffer
    while ((ftStatus == FT_OK) && (dwNumBytesToRead > 0))
    {
        if (dwNumBytesToRead > MAX_BUFFER)
            dwNumBytesToRead = MAX_BUFFER;
        FT_Read(ftHandle, &byInputBuffer, dwNumBytesToRead, &dwNumBytesRead);
        dwNumBytesToRead -= dwNumBytesRead;
    }
    //Read out the data from FT2232H receive buffer
//        ftStatus |= FT_SetUSBParameters(ftHandle, 65536, 65535);
    ftStatus |= FT_SetUSBParameters(ftHandle, MAX_BUFFER, MAX_BUFFER-1);
    //Set USB request transfer sizes to 1K
    ftStatus |= FT_SetChars(ftHandle, 0, 0, 0, 0);
    //Disable event and error characters
    ftStatus |= FT_SetTimeouts(ftHandle, 0, 200);
    //Sets the read and write timeouts in milliseconds
    ftStatus |= FT_SetLatencyTimer(ftHandle, 0);
    //Set the latency timer (default is 16mS)
    ftStatus |= FT_SetBitMode(ftHandle, 0x0, 0x00);
    //Reset controller
    ftStatus |= FT_SetBitMode(ftHandle, 0x0, FT_BITMODE_MPSSE);
    //Enable MPSSE mode
    if (ftStatus != FT_OK)
    {
        printf("Error in initializing the MPSSE %d\n", (int)ftStatus);
        FT_Close(ftHandle);
        return -1;
    }
    Sleep(50); // Wait for all the USB stuff to complete and work

    // -----------------------------------------------------------
    // At this point, the MPSSE is ready for commands
    // -----------------------------------------------------------

    if (!SynchorniseMPSSE ())
    {
        printf ("MPSSE synchronisation failed\n");
        return -1;
    }

    // -----------------------------------------------------------
    // Configure the MPSSE settings for BDM
    // Multple commands can be sent to the MPSSE with one FT_Write
    // -----------------------------------------------------------

    dwNumBytesToSend = 0; // Start with a fresh index
    // Set up the Hi-Speed specific commands for the FTx232H
    byOutputBuffer[dwNumBytesToSend++] = 0x8A;  // Use 60MHz master clock (disable divide by 5)
    byOutputBuffer[dwNumBytesToSend++] = 0x97;  // Turn off adaptive clocking (may be needed for ARM)
    byOutputBuffer[dwNumBytesToSend++] = 0x8D;  // Disable three-phase clocking
    ftStatus = FT_Write(ftHandle, byOutputBuffer, dwNumBytesToSend, &dwNumBytesSent);
    // Send off the HS-specific commands

    // Set initial states of the MPSSE interface - low byte, both pin directions and output values

    GPIO_ADBUS_data_direction = MASK_DSCK | MASK_DSDI | MASK_NEN245;  //  SRESET and HRESET are open collector, so only output when pulled low. DSDO is input always.
    GPIO_ADBUS_data_value     = MASK_SRESET | MASK_HRESET;
//    GPIO_ADBUS_data_value     = MASK_SRESET | MASK_HRESET;
    Set_GPIO_LowByte_output ();

    // Set initial states of the MPSSE interface - high byte, both pin directions and output values
    // Pin name Signal Direction    Config Initial State Config
    // ACBUS0   GPIOH0 input        0                       0
    // ACBUS1   GPIOH1 input        0                       0
    // ACBUS2   GPIOH2 input        0                       0
    // ACBUS3   GPIOH3 input        0                       0
    // ACBUS4   GPIOH4 input        0                       0
    // ACBUS5   GPIOH5 input        0                       0
    // ACBUS6   GPIOH6 input        0                       0
    // ACBUS7   GPIOH7 input        0                       0
    GPIO_ACBUS_data_direction = 0b00000000;
    GPIO_ACBUS_data_value     = 0b00000000;
    Set_GPIO_HighByte_output ();

  //  Set_ClockDivisor (dwClockDivisor);

    LoopBackOnOrOff (0);

    return 0;
}

int mpc8xx_bdm_init ()
{
    int index;

    index = find_FTDI_FT2232H_OR_FT4232H_A_index ();
    if (index>=0)
    {
        open_FT2232_for_BDM (index);
        return 0;
    }
    printf ("\nNo working  USB module FT2232H/FT4232H A found.\n");
    return -1;
}

void mpc8xx_bdm_reset (int hard, int soft)
{
    GPIO_ADBUS_data_value |= MASK_DSCK;
    GPIO_ADBUS_data_value &= ~MASK_DSDI;
    Set_GPIO_LowByte_output ();
    Sleep(3);

    if (soft)
        GPIO_ADBUS_data_value &= ~MASK_SRESET;  // always 0, act as open collector pull down, never push
    if (hard)
        GPIO_ADBUS_data_value &= ~MASK_HRESET;  // always 0, act as open collector pull down, never push
    if (soft)
        GPIO_ADBUS_data_direction |= MASK_SRESET; // output
    if (hard)
        GPIO_ADBUS_data_direction |= MASK_HRESET; // output

    Set_GPIO_LowByte_output ();
    Sleep(3);

    if (soft)
        GPIO_ADBUS_data_direction &= ~MASK_SRESET;  // input (simulate open collector)
    if (hard)
        GPIO_ADBUS_data_direction &= ~MASK_HRESET;  // input (simulate open collector)
    Set_GPIO_LowByte_output ();
    Sleep(3);

    GPIO_ADBUS_data_value &= ~MASK_DSCK;
    Set_GPIO_LowByte_output ();

    Sleep(3);
}

int mpc8xx_bdm_wait_power (int timeout)
{
    int n = 0;
    timeout *= 1000;
    while ((!mpc8xx_bdm_has_power ()) && (timeout--))
    {
        Sleep(1);
        n++;
        if (n>100)
        {
            printf ("p");
            n=0;
        }
    }
    if (timeout)
        return 0;
    return -1;
}

int mpc8xx_bdm_has_power ()
{
    return 1;
}

int mpc8xx_bdm_wait_freeze (int timeout)
{
    int n = 0;
    timeout *= 1000;
    while ((getFRZ() == 0) && timeout--)
    {
 //       printf ("waiting for FRZ (=low level) on FRZ%c", (char)13);
        if (n>20)
            Sleep(1);
        n++;
        if (n>100)
        {
            printf ("f");
            n=0;
        }
    }
    if (timeout)
        return 0;
    return -1;
}

int getRDY ()
{
    return inout1bit (0);
}

int mpc8xx_bdm_wait_ready (int timeout)
{
    int n = 0;
    timeout *= 1000;

    while (getRDY () && timeout--)
    {
        n++;
        if (n>100)
        {
            printf ("BDM data line not in READY state. All jumpers JP3, JP4 away from IC2, JP5, JP6 & JP7 mounted in adapter?\r");
            n=0;
        }
        if (n>20)
            Sleep(1);
    }
    if (timeout)
        return 0;
    return -1;
}

/*
static int lptbdm_clk_bit( int b )
{
	int res = 0;

	if(getDSDO ())
		res = 1;
    if (b)
        setDSDI ();
    else
        clearDSDI ();

    setDSCK ();
    clearDSCK ();

	return res;
}
*/

int mpc8xx_bdm_clk_serial (const bdm_in_t* in, bdm_out_t* out)
{
    uint8_t toPPC_first3bits;
    uint32_t toPPC_next32bits;
    uint8_t fromPPC_first3bits = 0;
    uint32_t fromPPC_next32bits = 0;
    int len, reslen=0, res=0;

	memset( out, 0 , sizeof(bdm_out_t) );
    if (verbose) printf ("%x.%08x ", (int)in->prefix, in->data);

    toPPC_first3bits = in->prefix;
    toPPC_next32bits = in->data;

//    if (toPPC_first3bits == 2)
//        printf ("check");

  	if (in->prefix & 2)
		len = 7;
	else
		len = 32;

    mpc8xx_bdm_wait_ready (30);

    inout34or9bits (toPPC_first3bits, toPPC_next32bits, len, &fromPPC_first3bits, &fromPPC_next32bits);

    out->data = fromPPC_next32bits;
    out->ready = (fromPPC_first3bits >> 2) & 1;
    out->status = fromPPC_first3bits & 3;

    switch (out->status)
    {
        case MPC8XX_BDM_STAT_CORE_DATA:
            reslen = 32;
            break;
        /* debug command while target not in debug mode */
        case MPC8XX_BDM_STAT_SEQ_ERROR:
            res = -3;
            reslen = 7;
            break;
        /* interrupt occurred, have to read ICR to clear */
        case MPC8XX_BDM_STAT_CORE_INTERRUPT:
            res = -4;
            reslen = 7;
            break;
        case MPC8XX_BDM_STAT_NULL:
            reslen = 7;
            break;
        default:
            return -5;
            break;
	}

    if (verbose)
        switch (reslen)
        {
        case 32:
            printf ("(%2d) %x.%08x (%2d) (%d) \n", len, (int)out->status, out->data, reslen, res);
            break;
        case 7:
            printf ("(%2d) %x.%02x (%2d) (%d) \n", len, (int)out->status, out->data &0xff, reslen, res);
            break;
        default:
            printf ("(%2d) %x.%04x (%2d) (%d) \n", len, (int)out->status, out->data, reslen, res);
            break;


        }
    return res;
}

/** Set the different timing values
 * @param delay_time
 * @param sleep_time
 * @param relaxed_timing
 * @return -1 on error else 0
 */
int mpc8xx_set_timing (unsigned int delay_time, unsigned int sleep_time, int relaxed_timing)
{
    return 0;
}
