/**
 * @file	FT2232BDM
 */

#ifndef __FT2232BDM_H__
#define __FT2232BDM_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdlib.h>
#include <windows.h>		// for windows specific keywords in ftd2xx.h

#define MASK_DSCK   0b00000001
#define MASK_DSDI   0b00000010
#define MASK_DSDO   0b00000100
#define MASK_SRESET 0b00010000
#define MASK_FRZ    0b00100000
#define MASK_HRESET 0b01000000
#define MASK_NEN245 0b10000000

#define MAX_BUFFER 1024

/** Data to be clocked into the target
 *
 */
struct bdm_in_s {
	unsigned int prefix	: 2;	/**< 2 bit prefix */
	unsigned int data;		/**< 32 bit data */
};

typedef struct bdm_in_s bdm_in_t;


/** Data that get shifted out of the target
 *
 */
struct bdm_out_s {
	unsigned int ready 	: 1;	/**< target ready status */
	unsigned int status 	: 2;	/**< target status */
	unsigned int data;		/**< target data */
};

typedef struct bdm_out_s bdm_out_t;

void set_verbose (int v);
int get_verbose ();

int show_FTDI_devices (int *numberOfUnitsOfWantedType, int list_of_wanted_units_indexes[]);
int GetGPIOinput (uint8_t HighByte);
int Get_GPIO_HighByte_input ();
int Get_GPIO_LowByte_input ();
void SetGPIO_output (uint8_t HighByte);
void Set_GPIO_HighByte_output ();
void Set_GPIO_LowByte_output ();
void LoopBackOnOrOff (int LoopBackOn);
void Set_ClockDivisor (DWORD dwClockDivisor);
int SynchorniseMPSSE ();
int get_ADBUS (uint8_t mask);
void set_clear_ADBUS (int value, uint8_t mask);
void setDSDI ();
void clearDSDI ();
void setDSCK ();
void clearDSCK ();
void setSRESET();
void clearSRESET ();
void setHRESET();
void clearHRESET();
int getFRZ ();
int getDSDO ();
int getRDY ();

void inout35bits (uint8_t out_first3bits, uint32_t out_next32bits, uint8_t *in__first3bits, uint32_t *in_next32bits);
int open_FT2232_for_BDM (int index);
int find_FTDI_FT2232H_Mini_Module_A_index ();

int mpc8xx_bdm_init();
void mpc8xx_bdm_reset (int hard, int soft);
//void mpc8xx_bdm_reset (void);
int mpc8xx_bdm_wait_power (int timeout);
int mpc8xx_bdm_has_power ();
int mpc8xx_bdm_wait_freeze (int timeout);
int mpc8xx_bdm_wait_ready (int timeout);
int mpc8xx_bdm_clk_serial (const bdm_in_t* in, bdm_out_t* out);

/** Set the different timing values
 * @param delay_time
 * @param sleep_time
 * @param relaxed_timing
 * @return -1 on error else 0
 */
int mpc8xx_set_timing (unsigned int delay_time, unsigned int sleep_time, int relaxed_timing);


#ifdef __cplusplus
}
#endif

#endif /* !__FT2232BDM__ */
