/*
 * (C) Copyright 2000
 * Paolo Scaffardi, AIRVENT SAM s.p.a - RIMINI(ITALY), arsenio@tin.it
 *
 * (C) Copyright 2000 Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Back ported to the 8xx platform (from the 8260 platform) by
 * Murray.Jensen@cmst.csiro.au, 27-Jan-01.
 */

#include <stdio.h>
#include <stdint.h>
#include <ctype.h>

#include "i2c.h"
#include "mpc8xxbdm.h"
#include "mpc8xxspr.h"

#ifndef	PROFF_IIC
#define PROFF_IIC	((unsigned int)0x0080)
#endif

/*-----------------------------------------------------------------------
 * Set default values
 */
#ifndef	CFG_I2C_SPEED
#define	CFG_I2C_SPEED	50000
#endif

#ifndef	CFG_I2C_SLAVE
#define	CFG_I2C_SLAVE	0xFE
#endif
/*-----------------------------------------------------------------------
 */

/* tx/rx timeout (we need the i2c early, so we don't use get_timer()) */
#define TOUT_LOOP 1000

#define NUM_RX_BDS 4
#define NUM_TX_BDS 4
#define MAX_TX_SPACE 256
#define I2C_RXTX_LEN 128	/* maximum tx/rx buffer length */

typedef struct I2C_BD
{
  unsigned short status;
  unsigned short length;
  unsigned char *addr;
} I2C_BD;

#define BD_OFFSET_STATUS 0
#define BD_OFFSET_LENGTH 2
#define BD_OFFSET_ADDR0 4
#define BD_OFFSET_ADDR1 6


#define BD_SC_EMPTY	((unsigned short)0x8000)	/* Recieve is empty */
#define BD_SC_READY	((unsigned short)0x8000)	/* Transmit is ready */
#define BD_SC_WRAP	((unsigned short)0x2000)	/* Last buffer descriptor */
#define BD_SC_INTRPT	((unsigned short)0x1000)	/* Interrupt on change */
#define BD_SC_LAST	((unsigned short)0x0800)	/* Last buffer in frame */
#define BD_SC_TC	((unsigned short)0x0400)	/* Transmit CRC */
#define BD_SC_CM	((unsigned short)0x0200)	/* Continous mode */
#define BD_SC_ID	((unsigned short)0x0100)	/* Rec'd too many idles */
#define BD_SC_P		((unsigned short)0x0100)	/* xmt preamble */
#define BD_SC_BR	((unsigned short)0x0020)	/* Break received */
#define BD_SC_FR	((unsigned short)0x0010)	/* Framing error */
#define BD_SC_PR	((unsigned short)0x0008)	/* Parity error */
#define BD_SC_OV	((unsigned short)0x0002)	/* Overrun */
#define BD_SC_CD	((unsigned short)0x0001)	/* Carrier Detect lost */

#define BD_I2C_TX_START 0x0400  /* special status for i2c: Start condition */

#define BD_I2C_TX_CL	0x0001	/* collision error */
#define BD_I2C_TX_UN	0x0002	/* underflow error */
#define BD_I2C_TX_NAK	0x0004	/* no acknowledge error */
#define BD_I2C_TX_ERR	(BD_I2C_TX_NAK|BD_I2C_TX_UN|BD_I2C_TX_CL)

#define BD_I2C_RX_ERR	BD_SC_OV

typedef void (*i2c_ecb_t)(int, int);	/* error callback function */

/* This structure keeps track of the bd and buffer space usage. */
typedef struct i2c_state {
	int		rx_idx;		/* index   to next free Rx BD */
	int		tx_idx;		/* index   to next free Tx BD */
	unsigned int rxbd;		/* pointer to next free Rx BD */
	unsigned int txbd;		/* pointer to next free Tx BD */
	int		tx_space;	/* number  of Tx bytes left   */
	unsigned int	tx_buf;	/* pointer to free Tx area    */
	i2c_ecb_t	err_cb;		/* error callback function    */
} i2c_state_t;

/* flags for i2c_send() and i2c_receive() */
#define I2CF_ENABLE_SECONDARY	0x01	/* secondary_address is valid		*/
#define I2CF_START_COND		0x02	/* tx: generate start condition		*/
#define I2CF_STOP_COND		0x04	/* tx: generate stop  condition		*/

/* return codes */
#define I2CERR_NO_BUFFERS	0x01	/* no more BDs or buffer space		*/
#define I2CERR_MSG_TOO_LONG	0x02	/* tried to send/receive to much data	*/
#define I2CERR_TIMEOUT		0x03	/* timeout in i2c_doio()		*/
#define I2CERR_QUEUE_EMPTY	0x04	/* i2c_doio called without send/receive */

/* error callback flags */
#define I2CECB_RX_ERR		0x10	/* this is a receive error		*/
#define     I2CECB_RX_ERR_OV	0x02	/* receive overrun error		*/
#define     I2CECB_RX_MASK	0x0f	/* mask for error bits			*/
#define I2CECB_TX_ERR		0x20	/* this is a transmit error		*/
#define     I2CECB_TX_CL	0x01	/* transmit collision error		*/
#define     I2CECB_TX_UN	0x02	/* transmit underflow error		*/
#define     I2CECB_TX_NAK	0x04	/* transmit no ack error		*/
#define     I2CECB_TX_MASK	0x0f	/* mask for error bits			*/
#define I2CECB_TIMEOUT		0x40	/* this is a timeout error		*/

#define DEBUG_I2C

#ifdef DEBUG_I2C
#define PRINTD(x) printf x
#else
#define PRINTD(x)
#endif

extern unsigned int g_immr;

unsigned int buf_in_target = 0x9000;

unsigned int base_DualPortRAM;
unsigned int base_ParameterRAM;
unsigned int I2C_base_param;

unsigned int rx_buffer_descriptor_DualPortAddress;
unsigned int tx_buffer_descriptor_DualPortAddress;
unsigned int tx_buffer_DualPortAddress;

unsigned int rx_buffer_descriptor_32bitAddress;
unsigned int tx_buffer_descriptor_32bitAddress;
unsigned int tx_buffer_32bitAddress;

int I2C_verbose = 1;

void show_I2C_parameter (uint32_t index, char *name, int nbytes)
{
    printf ("I2C parameter 0x%02x (%s) at abs 0x%08x is ", index, name, index+I2C_base_param);
    switch (nbytes)
    {
        case 1: printf ("0x%02x", mpc8xx_get_byte (I2C_base_param + index)); break;
        case 2: printf ("0x%04x", mpc8xx_get_halfword (I2C_base_param + index)); break;
        default: printf ("0x%08x", mpc8xx_get_word (I2C_base_param + index)); break;
    }
    printf ("\n");
}

void show_I2C_parameters ()
{
    show_I2C_parameter (0x00, "RBASE", 2);
    show_I2C_parameter (0x02, "TBASE", 2);
    show_I2C_parameter (0x04, "RFCR", 1);
    show_I2C_parameter (0x05, "TFCR", 1);
    show_I2C_parameter (0x06, "MRBLR", 2);
    show_I2C_parameter (0x08, "RSTATE", 4);
    show_I2C_parameter (0x0c, "RPTR", 4);
    show_I2C_parameter (0x10, "RBPTR", 2);
    show_I2C_parameter (0x12, "RCOUNT", 2);
    show_I2C_parameter (0x14, "RTEMP", 4);
    show_I2C_parameter (0x18, "TSTATE", 4);
    show_I2C_parameter (0x1c, "TPTR", 4);
    show_I2C_parameter (0x20, "TBPTR", 2);
    show_I2C_parameter (0x22, "TCOUNT", 2);
    show_I2C_parameter (0x24, "TTEMP", 4);

}

void show_portA_sprs ()
{
    mpc8xx_show_spr_s ("PBPAR");
    mpc8xx_show_spr_s ("PBDIR");
    mpc8xx_show_spr_s ("PBODR");
}

void show_I2C_register (char *name)
{
    printf ("I2C register %s at 0x%08x is 0x%02x\n", name, mpc8xx_spr_find_abs_address(name), mpc8xx_get_spr_s_byte (name));
}

void show_I2C_registers ()
{
    show_I2C_register ("I2MOD");
    show_I2C_register ("I2ADD");
    show_I2C_register ("I2BRG");
    show_I2C_register ("I2CER");
    show_I2C_register ("I2CMR");
    show_I2C_register ("I2COM");
    show_portA_sprs ();
    show_I2C_buffer_descriptors ();
}


/*
 * Returns the best value of I2BRG to meet desired clock speed of I2C with
 * input parameters (clock speed, filter, and predivider value).
 * It returns computer speed value and the difference between it and desired
 * speed.
 */
static inline int i2c_roundrate(int hz, int speed, int filter, int modval, int *brgval, int *totspeed)
{
    int moddiv = 1 << (5-(modval & 3)), brgdiv, div;

    PRINTD(("\t[I2C] trying hz=%d, speed=%d, filter=%d, modval=%d\n", hz, speed, filter, modval));

    div = moddiv * speed;
    brgdiv = (hz + div - 1) / div;

    PRINTD(("\t\tmoddiv=%d, brgdiv=%d\n", moddiv, brgdiv));

    *brgval = ((brgdiv + 1) / 2) - 3 - (2*filter);

    if ((*brgval < 0) || (*brgval > 255)) {
	  PRINTD(("\t\trejected brgval=%d\n", *brgval));
	  return -1;
    }

    brgdiv = 2 * (*brgval + 3 + (2 * filter));
    div = moddiv * brgdiv ;
    *totspeed = hz / div;

    PRINTD(("\t\taccepted brgval=%d, totspeed=%d\n", *brgval, *totspeed));

    return  0;
}

/*
 * Sets the I2C clock predivider and divider to meet required clock speed.
 */
static int i2c_setrate (int hz, int speed)
{
//	immap_t		*immap = (immap_t *) CFG_IMMR;
//	volatile i2c8xx_t *i2c = (i2c8xx_t *) & immap->im_i2c;
	int		brgval,
			modval,		/* 0-3 */
			bestspeed_diff = speed,
			bestspeed_brgval = 0,
			bestspeed_modval = 0,
			bestspeed_filter = 0,
			totspeed,
			filter = 0;	/* Use this fixed value */

	for (modval = 0; modval < 4; modval++) {
		if (i2c_roundrate(hz,speed,filter,modval,&brgval,&totspeed) == 0) {
			int diff = speed - totspeed;

			if ((diff >= 0) && (diff < bestspeed_diff)) {
				bestspeed_diff = diff;
				bestspeed_modval = modval;
				bestspeed_brgval = brgval;
				bestspeed_filter = filter;
			}
		}
	}

	PRINTD (("[I2C] Best is:\n"));
	PRINTD (("[I2C] CPU=%dhz RATE=%d F=%d I2MOD=%08x I2BRG=%08x DIFF=%dhz\n",
		hz,
		speed,
		bestspeed_filter,
		bestspeed_modval,
		bestspeed_brgval,
		bestspeed_diff));


//	i2c->i2c_i2mod |= ((bestspeed_modval & 3) << 1) | (bestspeed_filter << 3);
    mpc8xx_or_spr_s_byte ("I2MOD", ((bestspeed_modval & 3) << 1) | (bestspeed_filter << 3));
//	i2c->i2c_i2brg = bestspeed_brgval & 0xff;
    mpc8xx_set_spr_s_byte ("I2BRG",bestspeed_brgval & 0xff);

	PRINTD (("[I2C] i2mod=%08x i2brg=%08x\n", mpc8xx_get_spr_s_byte ("I2MOD"),mpc8xx_get_spr_s_byte ("I2BRG")));

	return 1;
}

void reset_cp ()
{
    #define mk_cr_cmd(CH, CMD)	((CMD << 8) | (CH << 4))
#define CPM_CR_CH_I2C	    ((unsigned short)0x0001)	/* I2C and IDMA1 */
#define CPM_CR_FLG	        ((unsigned short)0x0001)
#define CPM_CR_INIT_TRX		((unsigned short)0x0000)
#define CPM_CR_INIT_RX		((unsigned short)0x0001)
#define CPM_CR_INIT_TX		((unsigned short)0x0002)
#define CPM_CR_HUNT_MODE	((unsigned short)0x0003)
#define CPM_CR_STOP_TX		((unsigned short)0x0004)
#define CPM_CR_RESTART_TX	((unsigned short)0x0006)
#define CPM_CR_SET_GADDR	((unsigned short)0x0008)

//	cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TRX) | CPM_CR_FLG;
    mpc8xx_set_spr (0x9C0 | MPC8XX_SPRI_MASK, mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TRX) | CPM_CR_FLG);
    while (mpc8xx_get_spr (0x9C0 | MPC8XX_SPRI_MASK) & 1)
    {
        if (mpc8xx_nop () != 0)
            break;

        printf ("waiting for CP reset... ");
    }
}

void i2c_init(int speed, int slaveaddr)
{

    base_DualPortRAM = g_immr + 0x2000;
    base_ParameterRAM = base_DualPortRAM + 0x1C00;
	I2C_base_param = base_ParameterRAM + 0x80;

    rx_buffer_descriptor_DualPortAddress = 0x0;
    tx_buffer_descriptor_DualPortAddress = rx_buffer_descriptor_DualPortAddress + NUM_RX_BDS *sizeof(I2C_BD);
    tx_buffer_DualPortAddress = tx_buffer_descriptor_DualPortAddress + NUM_TX_BDS * sizeof(I2C_BD);

    rx_buffer_descriptor_32bitAddress = rx_buffer_descriptor_DualPortAddress + base_DualPortRAM;
    tx_buffer_descriptor_32bitAddress = tx_buffer_descriptor_DualPortAddress + base_DualPortRAM;
    tx_buffer_32bitAddress = tx_buffer_DualPortAddress + base_DualPortRAM;

	/* Initialize Port B I2C pins. */
//	cp->cp_pbpar |= 0x00000030;
    mpc8xx_or_spr_s ("PBPAR", 0x00000030);
//	cp->cp_pbdir |= 0x00000030;
    mpc8xx_or_spr_s ("PBDIR", 0x00000030);
//	cp->cp_pbodr |= 0x00000030;
    mpc8xx_or_spr_s ("PBODR", 0x00000030);



	/* Disable interrupts */
//	i2c->i2c_i2mod = 0x00;
    mpc8xx_set_spr_s_byte ("I2MOD", 0x00);
//	i2c->i2c_i2cmr = 0x00;
    mpc8xx_set_spr_s_byte ("I2COM", 0x00);
//	i2c->i2c_i2cer = 0xff;
    mpc8xx_set_spr_s_byte ("I2CER", 0xff);
//  i2c->i2c_i2add = slaveaddr;
    mpc8xx_set_spr_s_byte ("I2ADD", slaveaddr);

	/*
	 * Set the I2C BRG Clock division factor from desired i2c rate
	 * and current CPU rate (we assume sccr dfbgr field is 0;
	 * divide BRGCLK by 1)
	 */
	PRINTD(("[I2C] Setting rate...\n"));
	i2c_setrate (25000000, CFG_I2C_SPEED) ;

	/* Set I2C controller in master mode */
//	i2c->i2c_i2com = 0x01;
    mpc8xx_set_spr_s_byte ("I2COM", 0x01);

	/* Set SDMA bus arbitration level to 5 (SDCR) */
//	immap->im_siu_conf.sc_sdcr = 0x0001 ;
    mpc8xx_set_spr_s_byte ("SDCR", 0x0001);

	/* Initialize Tx/Rx parameters */
	mpc8xx_set_halfword (I2C_base_param + RBASE, rx_buffer_descriptor_DualPortAddress);
	mpc8xx_set_halfword (I2C_base_param + TBASE, tx_buffer_descriptor_DualPortAddress);
	mpc8xx_set_halfword (I2C_base_param + RBPTR, mpc8xx_get_halfword (I2C_base_param + RBASE));
	mpc8xx_set_halfword (I2C_base_param + TBPTR, mpc8xx_get_halfword (I2C_base_param + TBASE));

	PRINTD(("[I2C] rbase = %04x\n", mpc8xx_get_halfword (I2C_base_param + RBASE)));
	PRINTD(("[I2C] tbase = %04x\n", mpc8xx_get_halfword (I2C_base_param + TBASE)));
	PRINTD(("[I2C] rxbd = %08x\n", rx_buffer_descriptor_32bitAddress));
	PRINTD(("[I2C] txbd = %08x\n", tx_buffer_descriptor_32bitAddress));

	/* Set big endian byte order */
	mpc8xx_set_byte (I2C_base_param + TFCR, 0x10);
	mpc8xx_set_byte (I2C_base_param + RFCR, 0x10);

	/* Set maximum receive size. */
    mpc8xx_set_halfword (I2C_base_param + MRBLR, I2C_RXTX_LEN);

    mpc8xx_set_word (I2C_base_param + RSTATE, 0);
    mpc8xx_set_word (I2C_base_param + TSTATE, 0);

    reset_cp ();

	/* Clear events and interrupts */
//	i2c->i2c_i2cer = 0xff;
    mpc8xx_set_spr_s_byte ("I2CER", 0xff);
//	i2c->i2c_i2cmr = 0x00;
    mpc8xx_set_spr_s_byte ("I2CMR", 0x00);


    show_I2C_registers ();
    show_I2C_parameters ();
}


static void i2c_newio(i2c_state_t *state)
{
	PRINTD(("[I2C] i2c_newio\n"));

	state->rx_idx = 0;
	state->tx_idx = 0;
	state->rxbd = rx_buffer_descriptor_32bitAddress;
	state->txbd = tx_buffer_descriptor_32bitAddress;
	state->tx_space = MAX_TX_SPACE;
	state->tx_buf = tx_buffer_32bitAddress;
	state->err_cb = NULL;

	PRINTD(("[I2C] rxbd = %08x\n", state->rxbd));
	PRINTD(("[I2C] txbd = %08x\n",  state->txbd));
	PRINTD(("[I2C] tx_buf = %08x\n", state->tx_buf));

	/* clear the buffer memory */
//	memset((char *)state->tx_buf, 0, MAX_TX_SPACE);
	for (int i=0; i<MAX_TX_SPACE; i++)
        mpc8xx_set_byte (state->tx_buf + i, 0);

	PRINTD(("[I2C] tx_buf = %08x\n", state->tx_buf));
}

static int i2c_send(i2c_state_t *state,
	 unsigned char address,
	 unsigned char secondary_address,
	 unsigned int flags,
	 unsigned short size,
	 unsigned char *dataout)
{
//	volatile I2C_BD *txbd;
	int i,j;


	PRINTD(("[I2C] i2c_send add=%02d sec=%02d flag=%02d size=%d\n",
			address, secondary_address, flags, size));

	/* trying to send message larger than BD */
	if (size > I2C_RXTX_LEN)
	  return I2CERR_MSG_TOO_LONG;

	/* no more free bds */
	if (state->tx_idx >= NUM_TX_BDS || state->tx_space < (2 + size))
	  return I2CERR_NO_BUFFERS;

//	txbd = (I2C_BD *)state->txbd;
//	txbd->addr = state->tx_buf;

	PRINTD(("[I2C] txbd = %08x\n", tx_buffer_descriptor_32bitAddress));

	if (flags & I2CF_START_COND) {
		PRINTD(("[I2C] Formatting addresses...\n"));
		if (flags & I2CF_ENABLE_SECONDARY) {

//			txbd->length = size + 2;  /* Length of msg + dest addr */
            mpc8xx_set_halfword (state->txbd + BD_OFFSET_LENGTH, size + 2);
//			txbd->addr[0] = address << 1;
            mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR0, address << 1);
//			txbd->addr[1] = secondary_address;
            mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR1, secondary_address);
			i = 2;
		}
		else
        {
//		    txbd->length = size + 1;
            mpc8xx_set_halfword (state->txbd + BD_OFFSET_LENGTH, size+1);
//		    txbd->addr[0] = (address << 1);
            mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR0, (address << 1));
			i = 1;
		}
	} else
	{
//		txbd->length = size;  /* Length of message */
        mpc8xx_set_halfword (state->txbd + BD_OFFSET_LENGTH, size);
		i = 0;
	}

	/* set up txbd */
//	txbd->status = BD_SC_READY;
    mpc8xx_set_halfword (state->txbd + BD_OFFSET_STATUS, BD_SC_READY);

	if (flags & I2CF_START_COND)
//	    txbd->status |= BD_I2C_TX_START;
        mpc8xx_or_halfword (state->txbd + BD_OFFSET_STATUS, BD_I2C_TX_START);


	if (flags & I2CF_STOP_COND)
//	  txbd->status |= BD_SC_LAST | BD_SC_WRAP;
        mpc8xx_or_halfword (state->txbd + BD_OFFSET_STATUS, BD_SC_LAST | BD_SC_WRAP);

	/* Copy data to send into buffer */
	PRINTD(("[I2C] copy data...\n"));
	for(j = 0; j < size; i++, j++)
//	    txbd->addr[i] = dataout[j];
        mpc8xx_set_byte (mpc8xx_get_word (state->txbd + BD_OFFSET_ADDR0) + j, dataout[j]);

	PRINTD(("[I2C] txbd: length=0x%04x status=0x%04x addr[0]=0x%02x addr[1]=0x%02x\n",
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH),
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_STATUS),
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_ADDR0),
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_ADDR0)));

	/* advance state */
	state->tx_buf += mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH);
	state->tx_space -= mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH);
	state->tx_idx++;

//	state->txbd = (void*)(txbd + 1);

	return 0;
}

static int i2c_receive(i2c_state_t *state,
	    unsigned char address,
	    unsigned char secondary_address,
	    unsigned int flags,
	    unsigned short size_to_expect,
	    unsigned char *datain)
{
//	volatile I2C_BD *rxbd, *txbd;

	PRINTD(("[I2C] i2c_receive %02d %02d %02d\n", address, secondary_address, flags));

	/* Expected to receive too much */
	if (size_to_expect > I2C_RXTX_LEN)
	  return I2CERR_MSG_TOO_LONG;

	/* no more free bds */
	if (state->tx_idx >= NUM_TX_BDS || state->rx_idx >= NUM_RX_BDS || state->tx_space < 2)
	  return I2CERR_NO_BUFFERS;

//	rxbd = (I2C_BD *)state->rxbd;
//	txbd = (I2C_BD *)state->txbd;

	PRINTD(("[I2C] rxbd = %08x\n", rx_buffer_descriptor_32bitAddress));
	PRINTD(("[I2C] txbd = %08x\n", tx_buffer_descriptor_32bitAddress));

//	txbd->addr = state->tx_buf;
    mpc8xx_set_word (state->txbd + BD_OFFSET_ADDR0, state->tx_buf);

	/* set up TXBD for destination address */
	if (flags & I2CF_ENABLE_SECONDARY)
    {
//		txbd->length = 2;
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_LENGTH, 2);
//		txbd->addr[0] = address << 1;   /* Write data */
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR0, address << 1);
//		txbd->addr[1] = secondary_address;  /* Internal address */
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR1, secondary_address);
//		txbd->status = BD_SC_READY;
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_STATUS, BD_SC_READY);
	}
	else
    {
//		txbd->length = 1 + size_to_expect;
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_LENGTH, 1 + size_to_expect);
//		txbd->addr[0] = (address << 1) | 0x01;
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_ADDR0, (address << 1) | 0x01);
//		txbd->status = BD_SC_READY;
		mpc8xx_set_halfword (state->txbd + BD_OFFSET_STATUS, BD_SC_READY);

//		memset(&txbd->addr[1], 0, txbd->length);
        int length = mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH);
		for (int i=0; i<length; i++)
            mpc8xx_set_byte (state->txbd + BD_OFFSET_ADDR1 + i, 0);
	}


	/* set up rxbd for reception */

//	rxbd->status = BD_SC_EMPTY;
	mpc8xx_set_halfword (state->rxbd + BD_OFFSET_STATUS, BD_SC_EMPTY);

//	rxbd->length = size_to_expect;
	mpc8xx_set_halfword (state->rxbd + BD_OFFSET_LENGTH, size_to_expect);
//	rxbd->addr = datain;
	mpc8xx_set_word (state->rxbd + BD_OFFSET_ADDR0, (unsigned int)buf_in_target);

//	txbd->status |= BD_I2C_TX_START;
    mpc8xx_or_halfword (state->txbd + BD_OFFSET_STATUS, BD_I2C_TX_START);

	if (flags & I2CF_STOP_COND)
    {
//		txbd->status |= BD_SC_LAST | BD_SC_WRAP;
        mpc8xx_or_halfword (state->txbd + BD_OFFSET_STATUS, BD_SC_LAST | BD_SC_WRAP);
//		rxbd->status |= BD_SC_WRAP;
        mpc8xx_or_halfword (state->rxbd + BD_OFFSET_STATUS, BD_SC_WRAP);
	}

	PRINTD(("[I2C] txbd: length=0x%04x status=0x%04x addr[0]=0x%02x addr[1]=0x%02x\n",
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH), // txbd->length,
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_STATUS), // txbd->status,
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_ADDR0), //txbd->addr[0],
		   mpc8xx_get_halfword (state->txbd + BD_OFFSET_ADDR1))); // txbd->addr[1]));
	PRINTD(("[I2C] rxbd: length=0x%04x status=0x%04x addr[0]=0x%02x addr[1]=0x%02x\n",
		   mpc8xx_get_halfword (state->rxbd + BD_OFFSET_LENGTH), // rxbd->length,
		   mpc8xx_get_halfword (state->rxbd + BD_OFFSET_STATUS), // rxbd->status,
		   mpc8xx_get_halfword (state->rxbd + BD_OFFSET_ADDR0), //rxbd->addr[0],
		   mpc8xx_get_halfword (state->rxbd + BD_OFFSET_ADDR1))); // rxbd->addr[1]));

	/* advance state */
	state->tx_buf += mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH);
	state->tx_space -= mpc8xx_get_halfword (state->txbd + BD_OFFSET_LENGTH);
	state->tx_idx++;

//	state->txbd = (void*)(txbd + 1);
	state->rx_idx++;
//	state->rxbd = (void*)(rxbd + 1);

    for (int i=0; i< size_to_expect; i++)
	    *datain++ = mpc8xx_get_byte (buf_in_target+i);


	return 0;
}


void show_I2C_buffer_descriptors ()
{
    printf ("I2C TXBD_STATUS = 0x%04x LENGTH=0x%04x ADDR0=0x%04x ADDR1=0x%04x\n", mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS),
                                                      mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_LENGTH),
                                                      mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_ADDR0),
                                                      mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_ADDR1));
    printf ("I2C RXBD_STATUS = 0x%04x LENGTH=0x%04x ADDR0=0x%04x ADDR1=0x%04x\n", mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS),
                                                      mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_LENGTH),
                                                      mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_ADDR0),
                                                      mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_ADDR1));
}

static int i2c_doio(i2c_state_t *state)
{
	volatile int j = 0;
    int done = 0;
    unsigned int BDstatus;

	PRINTD(("[I2C] i2c_doio\n"));

	if (state->tx_idx <= 0 && state->rx_idx <= 0) {
		PRINTD(("[I2C] No I/O is queued\n"));
		return I2CERR_QUEUE_EMPTY;
	}

	mpc8xx_set_halfword (I2C_base_param + RBPTR, mpc8xx_get_halfword (I2C_base_param + RBASE));
	mpc8xx_set_halfword (I2C_base_param + TBPTR, mpc8xx_get_halfword (I2C_base_param + TBASE));

	/* Enable I2C */
	PRINTD(("[I2C] Enabling I2C...\n"));
    mpc8xx_or_spr_s_byte ("I2MOD", 0x01);

    show_I2C_registers ();
    show_I2C_parameters ();
//    show_I2C_buffer_descriptors ();

	/* Begin transmission */
    mpc8xx_or_spr_s_byte ("I2COM", 0x80);

    show_I2C_registers ();
    show_I2C_parameters ();
//    show_I2C_buffer_descriptors ();

	/* Loop until transmit & receive completed */
	if (state->tx_idx > 0)
    {
//		txbd = ((I2C_BD*)state->txbd) - 1;
		PRINTD(("[I2C] Transmitting...(txbd=0x%08x)\n", tx_buffer_descriptor_32bitAddress));

		done = 0;
		do
        {
            if (mpc8xx_nop () != 0)
               done = 1;

            BDstatus = mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS);
            if (!(BDstatus & BD_SC_READY))
                done = 1;

            if (j++ >= TOUT_LOOP)
                done = 1;

        } while (!done);
	}

    show_I2C_registers ();
    show_I2C_parameters ();
 //   show_I2C_buffer_descriptors ();

	if ((state->rx_idx > 0) && (j < TOUT_LOOP))
    {
//		rxbd = ((I2C_BD*)state->rxbd) - 1;
		PRINTD(("[I2C] Receiving...(rxbd=0x%08x)\n", rx_buffer_descriptor_32bitAddress));
		done = 0;
		do
        {
            if (mpc8xx_nop () != 0)
               done = 1;

            BDstatus = mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS);
            if (!(BDstatus & BD_SC_EMPTY))
                done = 1;

            if (j++ >= TOUT_LOOP)
                done = 1;

        } while (!done);
   }

	/* Turn off I2C */
    mpc8xx_and_spr_s_byte ("I2MOD", ~0x01);

	if (state->err_cb != NULL)
    {
		int n, i, b;

		/*
		 * if we have an error callback function, look at the
		 * error bits in the bd status and pass them back
		 */

		if ((n = state->tx_idx) > 0)
		{
			for (i = 0; i < n; i++)
            {
                mpc8xx_nop ();
                BDstatus = mpc8xx_get_halfword (tx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS);
 				if ((b = BDstatus & BD_I2C_TX_ERR) != 0)
					(*state->err_cb)(I2CECB_TX_ERR|b, i);
			}
		}

		if ((n = state->rx_idx) > 0)
        {
			for (i = 0; i < n; i++)
			{
                mpc8xx_nop ();
                BDstatus =   mpc8xx_get_halfword (rx_buffer_descriptor_32bitAddress + BD_OFFSET_STATUS);
				if ((b = BDstatus & BD_I2C_RX_ERR) != 0)
					(*state->err_cb)(I2CECB_RX_ERR|b, i);
			}
		}

		if (j >= TOUT_LOOP)
			(*state->err_cb)(I2CECB_TIMEOUT, 0);
	}

	return (j >= TOUT_LOOP) ? I2CERR_TIMEOUT : 0;
}

static int had_tx_nak;

static void i2c_test_callback(int flags, int xnum)
{
	if ((flags & I2CECB_TX_ERR) && (flags & I2CECB_TX_NAK))
		had_tx_nak = 1;
}

int i2c_probe(unsigned char chip)
{
	i2c_state_t state;
  	int rc;
	unsigned char buf[1];

	i2c_init(CFG_I2C_SPEED, CFG_I2C_SLAVE);

	i2c_newio(&state);

	state.err_cb = i2c_test_callback;
	had_tx_nak = 0;

	rc = i2c_receive(&state, chip, 0, I2CF_START_COND|I2CF_STOP_COND, 1, buf);

	if (rc != 0)
		return (rc);

	rc = i2c_doio(&state);

	if ((rc != 0) && (rc != I2CERR_TIMEOUT))
		return (rc);

	return (had_tx_nak);
}

int i2c_read(unsigned char chip, unsigned int addr, int alen, unsigned char *buffer, int len)
{
	i2c_state_t state;
	unsigned char xaddr[4];
	int rc;

	xaddr[0] = (addr >> 24) & 0xFF;
	xaddr[1] = (addr >> 16) & 0xFF;
	xaddr[2] = (addr >>  8) & 0xFF;
	xaddr[3] =  addr        & 0xFF;

	i2c_newio(&state);

	rc = i2c_send(&state, chip, 0, I2CF_START_COND, alen, &xaddr[4-alen]);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_read: i2c_send failed (%d)\n", rc);
		return 1;
	}

	rc = i2c_receive(&state, chip, 0, I2CF_STOP_COND, len, buffer);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_read: i2c_receive failed (%d)\n", rc);
		return 1;
	}

	rc = i2c_doio(&state);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_read: i2c_doio failed (%d)\n", rc);
		return 1;
	}
	return 0;
}

int i2c_write(unsigned char chip, unsigned int addr, int alen, unsigned char *buffer, int len)
{
	i2c_state_t state;
	unsigned char xaddr[4];
	int rc;

	xaddr[0] = (addr >> 24) & 0xFF;
	xaddr[1] = (addr >> 16) & 0xFF;
	xaddr[2] = (addr >>  8) & 0xFF;
	xaddr[3] =  addr        & 0xFF;

	i2c_newio(&state);

	rc = i2c_send(&state, chip, 0, I2CF_START_COND, alen, &xaddr[4-alen]);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_write: first i2c_send failed (%d)\n", rc);
		return 1;
	}

	rc = i2c_send(&state, 0, 0, I2CF_STOP_COND, len, buffer);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_write: second i2c_send failed (%d)\n", rc);
		return 1;
	}

	rc = i2c_doio(&state);
	if (rc != 0) {
		if (I2C_verbose)
			printf("i2c_write: i2c_doio failed (%d)\n", rc);
		return 1;
	}
	return 0;
}

unsigned char i2c_reg_read(unsigned char i2c_addr, unsigned char reg)
{
	unsigned char buf;

	i2c_init(CFG_I2C_SPEED, CFG_I2C_SLAVE);

	i2c_read(i2c_addr, reg, 1, &buf, 1);

	return (buf);
}

void i2c_reg_write(unsigned char i2c_addr, unsigned char reg, unsigned char val)
{
	i2c_init(CFG_I2C_SPEED, CFG_I2C_SLAVE);

	i2c_write(i2c_addr, reg, 1, &val, 1);
}
