#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "mpc8xxbdm.h"
#include "mpc8xxspr.h"
#include "mpc8xxI2C.h"

static int Last_value_written_to_SDA = -1;
static int Last_value_written_to_SCL = -1;

static unsigned char g_device_address_hi_bits = 0;

void I2C_SET_SDA_TO (int value)
{
    if (Last_value_written_to_SDA == value)
        return;

    if (value)
        mpc8xx_or_spr_s ("PBDAT", I2C_MASK_SDA);
    else
        mpc8xx_and_spr_s ("PBDAT", ~I2C_MASK_SDA);
    Last_value_written_to_SDA = value;
}

void I2C_SET_SCL_TO (int value)
{
    if (Last_value_written_to_SCL == value)
        return;

    if (value)
        mpc8xx_or_spr_s ("PBDAT", I2C_MASK_SCL);
    else
        mpc8xx_and_spr_s ("PBDAT", ~I2C_MASK_SCL);
    Last_value_written_to_SCL = value;
}

void I2C_START ()
{
    I2C_SET_SDA_TO (0);
    I2C_SET_SCL_TO (0);
}

void I2C_STOP ()
{
    I2C_SET_SCL_TO (1);
    I2C_SET_SDA_TO (1);
}

void I2C_SEND_BIT (int i)
{
    if (i)
        I2C_SET_SDA_TO (1);
    else
        I2C_SET_SDA_TO (0);

    I2C_SET_SCL_TO (1);
    I2C_SET_SCL_TO (0);
}

int I2C_GET_SDA ()
{
    int x;
    I2C_SET_SDA_TO (1);

    Last_value_written_to_SDA = -2;

    x  = mpc8xx_get_spr_s ("PBDAT") & I2C_MASK_SDA;
    if (x)
        return 1;
    else
        return 0;
}

int I2C_READ_BIT ()
{
    int bit;
    I2C_SET_SDA_TO (1);
    I2C_SET_SCL_TO (1);
    bit = I2C_GET_SDA ();
    I2C_SET_SCL_TO (0);
    return bit;
}

int I2C_GET_SCL ()
{
    int x;
    I2C_SET_SDA_TO (1);
    Last_value_written_to_SCL = -2;

    x = mpc8xx_get_spr_s ("PBDAT") & I2C_MASK_SCL;
    if (x)
        return 1;
    else
        return 0;
}

int I2C_SEND_BYTE (int x)
{
    int ACK;
    int i;

    for (i=0; i<8; i++)
    {
        I2C_SEND_BIT (x & 0x80);
        x <<= 1;
    }

    ACK = I2C_READ_BIT () ^ 1;

    if (!ACK)
        I2C_STOP ();
    return ACK;
}

void I2C_RECEIVE_BYTE (unsigned char *x)
{
    unsigned char b = 0;
    int i;

    for (i=0; i<8; i++)
    {
        b <<= 1;
        if (I2C_READ_BIT ())
            b |= 1;
    }

    *x = b;
}

int I2C_READ (unsigned char I2C_device_address, unsigned char *data, int N)
{
    int i, nreceived_OK = 0;;

    I2C_START ();
    if (I2C_SEND_BYTE ((g_device_address_hi_bits + I2C_device_address) | 1))
    {
        for (i=0; i<N; i++)
        {
            I2C_RECEIVE_BYTE (data++);
            nreceived_OK++;
            if (i == N-1)
                I2C_SEND_BIT (1); // NACK
            else
                I2C_SEND_BIT (0); // ACK
        }
        I2C_STOP ();
    }
    return nreceived_OK;
}

int I2C_WRITE (unsigned char I2C_device_address, unsigned char *data, int N)
{
    int i;

    for (i=0; i<N; i++)
    {
        if (!(I2C_SEND_BYTE (*data++)))
            break;
    }
    if (i == N)
        I2C_STOP ();

    return i;
}

int I2C_WRITE_AP (unsigned char I2C_device_address, unsigned char Register_Address, int WithStop)
{
    int r = 0;

    I2C_START ();
    if (I2C_SEND_BYTE ((g_device_address_hi_bits + I2C_device_address) & ~1))
    {
        if (I2C_SEND_BYTE (Register_Address))
        {
            if (WithStop)
                I2C_STOP ();
            r = 1;
        }
    }
    return r;
}


int I2C_read_buffer (unsigned char I2C_device_address, unsigned short address_in_device, unsigned char *buffer, int N)
{
    push_r0_r1 ();

    int current_preserve_r0 = preserve_r0;
    int current_preserve_r1 = preserve_r1;
    preserve_r0 = 0;
    preserve_r1 = 0;
    I2C_init ();

    g_device_address_hi_bits = ((address_in_device>>8) << 1);

    if (I2C_WRITE_AP (I2C_device_address, address_in_device & 0xff, 1))
        return I2C_READ (I2C_device_address, buffer, N);

    preserve_r0 = current_preserve_r0;
    preserve_r1 = current_preserve_r1;
    pop_r0_r1 ();
    return 0;
}

int I2C_write_buffer_max_4_bytes (unsigned char I2C_device_address, unsigned short address_in_device, unsigned char *buffer, int N)
{
    g_device_address_hi_bits = ((address_in_device>>8) << 1);

    if (I2C_WRITE_AP (I2C_device_address, address_in_device & 0xff, 0))
        return I2C_WRITE (I2C_device_address, buffer, N);
    return 0;
}

int I2C_write_buffer (unsigned char I2C_device_address, unsigned short address_in_device, unsigned char *buffer, int N)
{
    int current_preserve_r0 = preserve_r0;
    int current_preserve_r1 = preserve_r1;
    push_r0_r1 ();
    preserve_r0 = 0;
    preserve_r1 = 0;
    int total = 0;
    int NN, NNN;

    I2C_init ();

    while (total < N)
    {
        NN = N - total;
        if (NN>4)
            NN = 4;
        NNN = I2C_write_buffer_max_4_bytes (I2C_device_address, address_in_device, buffer, NN);
        address_in_device += NNN;
        buffer += NNN;
        total += NNN;
    }
    preserve_r0 = current_preserve_r0;
    preserve_r1 = current_preserve_r1;
    pop_r0_r1 ();
    return total;
}

void I2C_init ()
{
    mpc8xx_and_spr_s ("PBPAR", ~I2C_MASK);   // general purpose i/o pins
    mpc8xx_or_spr_s ("PBDIR", I2C_MASK);      // outputs
    mpc8xx_or_spr_s ("PBODR", I2C_MASK);      // open drain

    I2C_SEND_BYTE (0xff);
    I2C_STOP ();
}
